@@ -38,8 +38,6 @@
 #include "printf.h"
 
 
-#define VAR_ECHO_STYLE "ECHO_STYLE"
-
 /* type of format data used in the "printf" builtin */
 struct format_T {
     struct format_T *next;
@@ -165,7 +163,7 @@ int echo_builtin(int argc, void **argv)
 	fflush(stdout);
 
     if (ferror(stdout)) {
-	xerror(errno, Ngt("write error"));
+	xerror(errno, Ngt("cannot print to standard output"));
 	return Exit_FAILURE;
     }
     return Exit_SUCCESS;
@@ -288,8 +286,7 @@ int printf_builtin(int argc, void **argv)
     while ((opt = xgetopt_long(argv, L"+", help_option, NULL))) {
 	switch (opt) {
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:  print_usage:
 		fprintf(stderr, gt("Usage:  printf format [value...]\n"));
 		return Exit_ERROR;
@@ -318,7 +315,7 @@ int printf_builtin(int argc, void **argv)
 end:
 	fflush(stdout);
 	if (ferror(stdout)) {
-	    xerror(errno, Ngt("write error"));
+	    xerror(errno, Ngt("cannot print to standard output"));
 	    printf_error = true;
 	}
     } else {
