@@ -79,8 +79,8 @@ enum srchcmdtype_T {
     sct_builtin  = 1 << 1,  /* search builtins */
     sct_function = 1 << 2,  /* search functions */
     sct_defpath  = 1 << 3,  /* search the default PATH */
-    sct_argc1    = 1 << 4,  /* argc == 1 */
-    sct_rbpath   = 1 << 5,  /* return path of regular builtin */
+    sct_rbpath   = 1 << 4,  /* return path of regular builtin */
+    sct_noslash  = 1 << 5,  /* don't check if name contains slash */
 };
 
 /* info about a simple command to execute */
@@ -124,16 +124,19 @@ static inline void connect_pipes(pipeinfo_T *pi)
     __attribute__((nonnull));
 
 static void exec_commands(command_T *c, exec_T type);
-static pid_t exec_process(command_T *c, exec_T type, pipeinfo_T *pi, pid_t pgid)
+static pid_t exec_process(
+	command_T *restrict c, exec_T type, pipeinfo_T *restrict pi, pid_t pgid)
     __attribute__((nonnull));
 static void search_command(
 	const char *restrict name, const wchar_t *restrict wname,
 	commandinfo_T *restrict ci, enum srchcmdtype_T type)
     __attribute__((nonnull));
+static inline bool is_special_builtin(const char *cmdname)
+    __attribute__((nonnull,pure));
 static inline bool assignment_is_temporary(enum cmdtype_T type)
     __attribute__((const));
-static bool including_path_assignment(const assign_T *a)
-    __attribute__((pure));
+static bool command_not_found_handler(void *const *argv)
+    __attribute__((nonnull));
 static void exec_nonsimple_command(command_T *c, bool finally_exit)
     __attribute__((nonnull));
 static void exec_simple_command(const commandinfo_T *ci,
@@ -151,6 +154,9 @@ static inline int xexecve(
 	const char *path, char *const *argv, char *const *envp)
     __attribute__((nonnull(1)));
 
+static int exec_iteration(void *const *commands, const char *codename)
+    __attribute__((nonnull));
+
 
 /* exit status of the last command */
 int laststatus = Exit_SUCCESS;
@@ -178,6 +184,12 @@ static struct execinfo {
 /* Note that `execinfo' is not reset when a subshell forks. */
 #define INIT_EXECINFO { 0, 0, ee_none }  /* used to initialize `execinfo' */
 
+/* state of currently executed iteration. */
+static struct iterinfo {
+    bool iterating;
+    enum { ie_none, ie_continue, ie_break } exception;
+} iterinfo;
+
 /* This flag is set while the "exec" builtin is executed. */
 static bool exec_builtin_executed = false;
 
@@ -188,16 +200,14 @@ static assign_T *last_assign;
 static xwcsbuf_T xtrace_buffer;
 
 
-/* Returns true iff we're breaking/continuing/returning now. */
-bool need_break(void)
-{
-    return execinfo.breakcount > 0 || execinfo.exception != ee_none;
-}
-
 /* Resets `execinfo' to the initial state. */
 void reset_execinfo(void)
 {
-    execinfo = (struct execinfo) INIT_EXECINFO;
+    execinfo = (struct execinfo) {
+	.loopnest = 0,
+	.breakcount = 0,
+	.exception = ee_none,
+    };
 }
 
 /* Saves the current `execinfo' and returns it.
@@ -206,7 +216,7 @@ struct execinfo *save_execinfo(void)
 {
     struct execinfo *save = xmalloc(sizeof execinfo);
     *save = execinfo;
-    execinfo = (struct execinfo) INIT_EXECINFO;
+    reset_execinfo();
     return save;
 }
 
@@ -217,10 +227,16 @@ void load_execinfo(struct execinfo *save)
     free(save);
 }
 
-/* Returns true iff `ee_return' is pending. */
-bool return_pending(void)
+/* Returns true iff `parse_and_exec' should quit execution. */
+inline bool return_pending(void)
 {
-    return execinfo.exception == ee_return;
+    return execinfo.exception != ee_none || iterinfo.exception != ie_none;
+}
+
+/* Returns true iff we're breaking/continuing/returning now. */
+bool need_break(void)
+{
+    return execinfo.breakcount > 0 || return_pending();
 }
 
 
@@ -616,87 +632,46 @@ done:
  * If the command was executed without forking, `laststatus' is set and
  * 0 is returned.
  * if `type' is `execself', this function never returns. */
-pid_t exec_process(command_T *c, exec_T type, pipeinfo_T *pi, pid_t pgid)
+pid_t exec_process(
+	command_T *restrict c, exec_T type, pipeinfo_T *restrict pi, pid_t pgid)
 {
     bool early_fork;   /* do early fork? */
-    bool later_fork;   /* do later fork? */
     bool finally_exit; /* never return? */
     int argc;
     void **argv = NULL;
     char *argv0 = NULL;
-    commandinfo_T cmdinfo = cmdinfo;
+    pid_t cpid = 0;
 
     current_lineno = c->c_lineno;
 
-    /* early fork if `type' is `execasync' or there is a pipe. */
-    early_fork = (type != execself) && (type == execasync
-	|| pi->pi_fromprevfd >= 0 || pi->pi_tonextfds[PIDX_OUT] >= 0);
+    /* do early fork if `type' is `execasync', the command type is subshell,
+     * or there is a pipe. */
+    early_fork = (type != execself)
+	&& (type == execasync || c->c_type == CT_SUBSHELL
+		|| pi->pi_fromprevfd >= 0 || pi->pi_tonextfds[PIDX_OUT] >= 0);
+    finally_exit = (type == execself);
     if (early_fork) {
 	sigtype_T sigtype = (type == execasync) ? t_quitint : 0;
-	pid_t cpid = fork_and_reset(pgid, type == execnormal, sigtype);
+	cpid = fork_and_reset(pgid, type == execnormal, sigtype);
 	if (cpid != 0)
-	    return cpid;
+	    goto done;
+	finally_exit = true;
     }
 
-    switch (c->c_type) {
-    case CT_SIMPLE:
-	if (!expand_line(c->c_words, &argc, &argv))
-	    goto exp_fail;
+    if (c->c_type == CT_SIMPLE) {
+	if (!expand_line(c->c_words, &argc, &argv)) {
+	    laststatus = Exit_EXPERROR;
+	    goto done;
+	}
 	if (argc == 0) {
-	    later_fork = finally_exit = false;
 	    argv0 = NULL;
 	} else {
 	    argv0 = malloc_wcstombs(argv[0]);
 	    if (!argv0)
 		argv0 = xstrdup("");
-	    search_command(argv0, argv[0], &cmdinfo,
-		    sct_external | sct_builtin | sct_function
-			| (argc == 1 ? sct_argc1 : 0));
-
-	    /* fork for an external command.
-	     * don't fork for a built-in or a function */
-	    later_fork = finally_exit = (cmdinfo.type == externalprogram);
-
-	    /* if command isn't found and there's no redirection or assignment,
-	     * we're done. */
-	    if (cmdinfo.type == externalprogram && !cmdinfo.ci_path
-		    && !c->c_redirs && !c->c_assigns) {
-		xerror(0, Ngt("%s: no such command or function"), argv0);
-		laststatus = Exit_NOTFOUND;
-		recfree(argv, free);
-		free(argv0);
-		goto done;
-	    }
-	}
-	break;
-    case CT_SUBSHELL:
-	later_fork = finally_exit = true;
-	break;
-    default:
-	later_fork = finally_exit = false;
-	break;
-    }
-    /* `argc', `argv' and `cmdinfo' are used only for `CT_SIMPLE'. */
-
-    if (early_fork || type == execself)
-	later_fork = false, finally_exit = true;
-    assert(!(early_fork && later_fork));  /* don't fork twice */
-    assert(!(early_fork || later_fork) || finally_exit);  /* exit if fork */
-
-    bool ext = (c->c_type == CT_SIMPLE) && (argc > 0)
-		&& (cmdinfo.type == externalprogram);
-    if (later_fork) {
-	pid_t cpid = fork_and_reset(pgid, type == execnormal, ext ? t_leave :0);
-	if (cpid != 0) {
-	    recfree(argv, free);
-	    free(argv0);
-	    return cpid;
 	}
     }
-    if (ext) {
-	restore_job_signals();
-	restore_interactive_signals();
-    }
+    /* `argc' and `argv' are used only for `CT_SIMPLE'. */
 
     lastcmdsubstatus = Exit_SUCCESS;
 
@@ -704,74 +679,116 @@ pid_t exec_process(command_T *c, exec_T type, pipeinfo_T *pi, pid_t pgid)
     connect_pipes(pi);
 
     /* open redirections */
-    savefd_T *savefd;
-    if (!open_redirections(c->c_redirs, finally_exit ? NULL : &savefd))
-	goto redir_fail;
+    savefd_T *savefd = NULL;
+    if (!open_redirections(c->c_redirs, finally_exit ? NULL : &savefd)) {
+	/* On redirection error, the command is not executed. */
+	laststatus = Exit_REDIRERR;
+	if (posixly_correct && !is_interactive_now && c->c_type == CT_SIMPLE &&
+		argc > 0 && is_special_builtin(argv0))
+	    finally_exit = true;
+	goto done2;
+    }
     
     /* redirect stdin to "/dev/null" if non-interactive and asynchronous */
     if (type == execasync && pi->pi_fromprevfd < 0)
 	maybe_redirect_stdin_to_devnull();
 
-    /* execute! */
-    if (c->c_type == CT_SIMPLE) {
-	last_assign = c->c_assigns;
-	if (argc == 0) {
-	    if (do_assignments(c->c_assigns, false, shopt_allexport))
-		laststatus = lastcmdsubstatus;
-	    else
-		laststatus = Exit_ASSGNERR;
-	    print_xtrace(NULL);
-	} else {
-	    bool temp = c->c_assigns && assignment_is_temporary(cmdinfo.type);
-	    if (temp)
-		open_new_environment(true);
-	    if (do_assignments(c->c_assigns, temp, true)) {
-		if (including_path_assignment(c->c_assigns))
-		    search_command(argv0, argv[0], &cmdinfo,
-			    sct_external | sct_builtin | sct_function
-				| (argc == 1 ? sct_argc1 : 0));
-		exec_simple_command(&cmdinfo, argc, argv0, argv, finally_exit);
-	    } else {
-		print_xtrace(NULL);
-		laststatus = Exit_ASSGNERR;
-		if (!is_interactive && cmdinfo.type == specialbuiltin)
-		    finally_exit = true;
-	    }
-	    if (temp)
-		close_current_environment();
-	}
-	recfree(argv, free);
-	free(argv0);
-    } else {
+    if (c->c_type != CT_SIMPLE) {
 	exec_nonsimple_command(c, finally_exit);
+	goto done1;
     }
-    if (finally_exit)
-	exit_shell();
 
-    if (exec_builtin_executed && laststatus == Exit_SUCCESS)
+    last_assign = c->c_assigns;
+    if (argc == 0) {
+	/* if there is no command word, just perform assignments */
+	if (do_assignments(c->c_assigns, false, shopt_allexport))
+	    laststatus = lastcmdsubstatus;
+	else
+	    laststatus = Exit_ASSGNERR;
+	print_xtrace(NULL);
+	goto done2;
+    }
+
+    commandinfo_T cmdinfo;
+    bool temp;
+
+    /* First, we check if the command is a special builtin or a function
+     * and determine whether we have to open a temporary environment. */
+    search_command(argv0, argv[0], &cmdinfo,
+	    sct_builtin | sct_function);
+    temp = c->c_assigns && assignment_is_temporary(cmdinfo.type);
+    if (temp)
+	open_new_environment(true);
+
+    /* perform assignments */
+    if (!do_assignments(c->c_assigns, temp, true)) {
+	/* On assignment error, the command is not executed. */
+	print_xtrace(NULL);
+	laststatus = Exit_ASSGNERR;
+	if (posixly_correct && !is_interactive &&
+		cmdinfo.type == specialbuiltin)
+	    finally_exit = true;
+	goto done3;
+    }
+
+    /* find command path */
+    if (cmdinfo.type == externalprogram) {
+	if (posixly_correct) {
+	    search_command(argv0, argv[0], &cmdinfo,
+		    sct_external | sct_builtin);
+	} else {
+	    if (wcschr(argv[0], L'/')) {
+		cmdinfo.ci_path = argv0;
+		if (!is_executable_regular(argv0))
+		    if (command_not_found_handler(argv))
+			goto done3;
+	    } else {
+		search_command(argv0, argv[0], &cmdinfo,
+			sct_external | sct_builtin | sct_noslash);
+		if (cmdinfo.type == externalprogram && cmdinfo.ci_path == NULL)
+		    if (command_not_found_handler(argv))
+			goto done3;
+	    }
+	    /* If the command was not found and the shell is not in POSIXly
+	     * correct mode, the not-found handler is called. */
+	}
+    }
+
+    /* create a child process to execute the external command */
+    if (cmdinfo.type == externalprogram && !finally_exit) {
+	cpid = fork_and_reset(pgid, type == execnormal, t_leave);
+	if (cpid != 0)
+	    goto done3;
+	finally_exit = true;
+    }
+
+    /* execute! */
+    exec_simple_command(&cmdinfo, argc, argv0, argv, finally_exit);
+
+    /* Redirections are not undone after a successful "exec" command:
+     * remove the saved data of file descriptors. */
+    if (exec_builtin_executed && laststatus == Exit_SUCCESS) {
 	clear_savefd(savefd);
-    else
-	undo_redirections(savefd);
+	savefd = NULL;
+    }
     exec_builtin_executed = false;
-    return 0;
 
-exp_fail:
-    laststatus = Exit_EXPERROR;
-done:
-    if (early_fork || type == execself)
-	exit_shell();
-    return 0;
-redir_fail:
-    laststatus = Exit_REDIRERR;
-    if (finally_exit)
-	exit_shell();
-    if (posixly_correct && !is_interactive_now && c->c_type == CT_SIMPLE &&
-	    argc > 0 && cmdinfo.type == specialbuiltin)
-	exit_shell();
-    undo_redirections(savefd);
+done3:
+    if (temp)
+	close_current_environment();
+done2:
     if (c->c_type == CT_SIMPLE)
 	recfree(argv, free), free(argv0);
-    return 0;
+done1:
+    undo_redirections(savefd);
+done:
+    if (finally_exit)
+	exit_shell();
+    if (cpid < 0) {
+	laststatus = Exit_NOEXEC;
+	cpid = 0;
+    }
+    return cpid;
 }
 
 /* Forks a subshell and does some settings.
@@ -856,28 +873,32 @@ pid_t fork_and_reset(pid_t pgid, bool fg, sigtype_T sigtype)
  * either a function pointer to the builtin (if the `sct_rbpath' flag is not
  * specified in `type') or the path of the corresponding external command
  * (otherwise). */
+/* In the POSIXly correct mode, you need the sct_external flag to find a regular
+ * builtin. */
 void search_command(
 	const char *restrict name, const wchar_t *restrict wname,
 	commandinfo_T *restrict ci, enum srchcmdtype_T type)
 {
-    if (wcschr(wname, L'/')) {
+    const builtin_T *bi;
+    command_T *funcbody;
+
+    if (!(type & sct_noslash) && wcschr(wname, L'/')) {
 	if (!(type & sct_external))
 	    goto notfound;
-	if (shopt_autocd && (type & sct_argc1) && is_directory(name))
-	    goto autocd;
 	ci->type = externalprogram;
 	ci->ci_path = name;
 	return;
     }
 
     /* search builtins and functions. */
-    const builtin_T *bi = (type & sct_builtin) ? get_builtin(name) : NULL;
-    command_T *funcbody = (type & sct_function) ? get_function(wname) : NULL;
+    bi = (type & sct_builtin) ? get_builtin(name) : NULL;
     if (bi && bi->type == BI_SPECIAL) {
 	ci->type = specialbuiltin;
 	ci->ci_builtin = bi->body;
 	return;
-    } else if (funcbody) {
+    }
+    funcbody = (type & sct_function) ? get_function(wname) : NULL;
+    if (funcbody) {
 	ci->type = function;
 	ci->ci_function = funcbody;
 	return;
@@ -891,7 +912,7 @@ void search_command(
 	return;
     }
 
-    if (!(type & sct_external))
+    if (!(type & (sct_external | sct_rbpath)))
 	goto notfound;
 
     ci->ci_path = (type & sct_defpath)
@@ -901,56 +922,83 @@ void search_command(
 	ci->type = regularbuiltin;
 	if (!(type & sct_rbpath))
 	    ci->ci_builtin = bi->body;
-    } else if (!ci->ci_path && shopt_autocd && (type & sct_argc1)
-	    && is_directory(name)) {
-	goto autocd;
     } else {
 	ci->type = externalprogram;
+	if (!(type & sct_external))
+	    ci->ci_path = NULL;
     }
     return;
 
-autocd:
-    ci->type = semispecialbuiltin;
-    ci->ci_builtin = cd_builtin;
-    return;
 notfound:
     ci->type = externalprogram;
     ci->ci_path = NULL;
     return;
 }
 
+/* Returns true iff the specified command is a special builtin. */
+bool is_special_builtin(const char *cmdname)
+{
+    const builtin_T *bi = get_builtin(cmdname);
+    return bi && bi->type == BI_SPECIAL;
+}
+
 /* Determines whether the assignments should be temporary according to `type'.*/
 bool assignment_is_temporary(enum cmdtype_T type)
 {
     switch (type) {
-	/* For external programs the result should be true in theory, but
-	 * actually we don't have to make the variables temporary because
-	 * the process forks. */
-	case externalprogram:
 	case specialbuiltin:
 	case function:
 	    return false;
 	case semispecialbuiltin:
 	case regularbuiltin:
+	case externalprogram:
 	    return true;
 	default:
 	    assert(false);
     }
 }
 
-/* Checks if the specified assignments include an assignment to $PATH. */
-bool including_path_assignment(const assign_T *a)
+/* Executes $COMMAND_NOT_FOUND_HANDLER if any.
+ * Returns true if the hander is executed and $HANDLED is set non-empty.
+ * Returns false otherwise. */
+bool command_not_found_handler(void *const *argv)
 {
-    while (a) {
-	if (wcscmp(a->a_name, L VAR_PATH) == 0)
-	    return true;
-	a = a->next;
+    static bool handling = false;
+    if (handling)
+	return false;  /* don't allow reentrance */
+    handling = true;
+
+    bool handled;
+    int result;
+
+    open_new_environment(false);
+    set_positional_parameters(argv);
+    set_variable(L VAR_HANDLED, xwcsdup(L""), SCOPE_LOCAL, false);
+
+    result = exec_variable_as_commands(
+	    L VAR_COMMAND_NOT_FOUND_HANDLER, VAR_COMMAND_NOT_FOUND_HANDLER);
+    if (result >= 0) {
+	const wchar_t *handledv = getvar(L VAR_HANDLED);
+	handled = (handledv != NULL && handledv[0] != L'\0');
+    } else {
+	handled = false;
+    }
+
+    close_current_environment();
+
+    handling = false;
+    if (handled) {
+	laststatus = result;
+	return true;
+    } else {
+	return false;
     }
-    return false;
 }
 
 /* Executes a command whose type is not `CT_SIMPLE'.
- * The redirections for the command is not performed in this function. */
+ * The redirections for the command is not performed in this function.
+ * For CT_SUBSHELL, this function must be called in an already-forked subshell.
+ */
 void exec_nonsimple_command(command_T *c, bool finally_exit)
 {
     c = comsdup(c);
@@ -991,16 +1039,17 @@ void exec_simple_command(
 	const commandinfo_T *ci, int argc, char *argv0, void **argv,
 	bool finally_exit)
 {
-    assert(argv[argc] == NULL);
+    assert(plcount(argv) == (size_t) argc);
 
     print_xtrace(argv);
     switch (ci->type) {
     case externalprogram:
-	assert(finally_exit || !ci->ci_path);
 	if (ci->ci_path == NULL) {
 	    xerror(0, Ngt("%s: no such command or function"), argv0);
 	    laststatus = Exit_NOTFOUND;
 	} else {
+	    assert(finally_exit);
+
 	    char *mbsargv[argc + 1];
 	    mbsargv[0] = argv0;
 	    for (int i = 1; i < argc; i++) {
@@ -1263,6 +1312,68 @@ wchar_t *exec_command_substitution(const wchar_t *code)
     }
 }
 
+/* Executes commands in the given array of wide strings (iterative execution).
+ * The strings are parsed and executed one by one.
+ * If the iteration is interrupted by the "break -i" command, the remaining
+ * elements are not executed.
+ * `codename' is passed to `exec_wcs' as the command name.
+ * If `saveparsestate' is true, the parser's state is saved during execution.
+ * Returns the exit status of the executed command (or zero if none executed).
+ * When this function returns, `laststatus' is restored to the original value.*/
+int exec_iteration(void *const *commands, const char *codename)
+{
+    int savelaststatus = laststatus, commandstatus = Exit_SUCCESS;
+    struct iterinfo saveiterinfo = iterinfo;
+
+    for (void *const *c = commands; *c; c++) {
+	iterinfo.iterating = true;
+	iterinfo.exception = ie_none;
+	exec_wcs(*c, codename, false);
+	commandstatus = laststatus;
+	laststatus = savelaststatus;
+	if (iterinfo.exception == ie_break)
+	    break;
+    }
+    iterinfo = saveiterinfo;
+    return commandstatus;
+}
+
+/* Executes the value of the specified variable.
+ * The variable value is parsed as a command string.
+ * If the `varname' names an array, every element of the array is executed (but
+ * if the iteration is interrupted by the "break -i" command, the remaining
+ * elements are not executed).
+ * `codename' is passed to `exec_wcs' as the command name.
+ * Returns the exit status of the executed command (or zero if none executed or
+ * -1 if the variable is unset).
+ * When this function returns, `laststatus' is restored to the original value.*/
+int exec_variable_as_commands(const wchar_t *varname, const char *codename)
+{
+    struct get_variable gv = get_variable(varname);
+    void **array;
+
+    switch (gv.type) {
+	case GV_NOTFOUND:
+	    return -1;
+	case GV_SCALAR:
+	    array = gv.values;
+	    break;
+	case GV_ARRAY:
+	    /* copy the array values in case they are unset during execution */
+	    array = duparrayn(gv.values, gv.count, copyaswcs);
+	    break;
+	case GV_ARRAY_CONCAT:
+	    /* should execute the concatenated value, but not supported now */
+	    return -1;
+	default:
+	    assert(false);
+    }
+
+    int result = exec_iteration(array, codename);
+    recfree(array, free);
+    return result;
+}
+
 
 /********** Builtins **********/
 
@@ -1272,12 +1383,19 @@ static int command_builtin_execute(
 	int argc, void **argv, enum srchcmdtype_T type)
     __attribute__((nonnull));
 static bool print_command_info(const wchar_t *commandname,
-	enum srchcmdtype_T type, bool humanfriendly)
+	enum srchcmdtype_T type, bool keywordalias, bool humanfriendly)
     __attribute__((nonnull));
 static void print_command_absolute_path(
 	const char *name, const char *path, bool humanfriendly)
     __attribute__((nonnull));
 
+/* Options for the "break", "continue" and "eval" builtins. */
+static const struct xoption iter_options[] = {
+    { L"iteration", xno_argument, L'i', },
+    { L"help",      xno_argument, L'-', },
+    { NULL, 0, 0, },
+};
+
 /* "return" builtin */
 int return_builtin(int argc, void **argv)
 {
@@ -1287,8 +1405,7 @@ int return_builtin(int argc, void **argv)
     while ((opt = xgetopt_long(argv, L"", help_option, NULL))) {
 	switch (opt) {
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:  print_usage:
 		fprintf(stderr, gt("Usage:  %ls [n]\n"), ARGV(0));
 		SPECIAL_BI_ERROR;
@@ -1300,12 +1417,16 @@ int return_builtin(int argc, void **argv)
 
     int status;
     const wchar_t *statusstr = ARGV(xoptind);
-    if (statusstr != NULL
-	    && xwcstoi(statusstr, 10, &status)
-	    && status >= 0) {
-	status &= 0xFF;
+    if (statusstr != NULL) {
+	if (xwcstoi(statusstr, 10, &status) && status >= 0) {
+	    status &= 0xFF;
+	} else {
+	    xerror(0, Ngt("`%ls' is not a valid integer"), statusstr);
+	    status = Exit_ERROR;
+	    SPECIAL_BI_ERROR;
+	    /* return anyway */
+	}
     } else {
-	/* ignore `statusstr' if not a valid non-negative integer */
 	status = (savelaststatus >= 0) ? savelaststatus : laststatus;
     }
     execinfo.exception = ee_return;
@@ -1324,102 +1445,139 @@ const char return_help[] = Ngt(
 int break_builtin(int argc, void **argv)
 {
     wchar_t opt;
+    bool iter = false;
 
     xoptind = 0, xopterr = true;
-    while ((opt = xgetopt_long(argv, L"", help_option, NULL))) {
+    while ((opt = xgetopt_long(argv, L"i", iter_options, NULL))) {
 	switch (opt) {
+	    case L'i':
+		iter = true;
+		break;
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:  print_usage:
-		fprintf(stderr, gt("Usage:  %ls [n]\n"), ARGV(0));
+		fprintf(stderr,
+			gt("Usage:  %ls [n]\n        %ls -i\n"),
+			ARGV(0), ARGV(0));
 		SPECIAL_BI_ERROR;
 		return Exit_ERROR;
 	}
     }
-    if (argc - xoptind > 1)
+    if (argc - xoptind > (iter ? 0 : 1))
 	goto print_usage;
 
-    unsigned count;
-    const wchar_t *countstr = ARGV(xoptind);
-    if (countstr == NULL) {
-	count = 1;
-    } else {
-	unsigned long countl;
-	if (!xwcstoul(countstr, 0, &countl)) {
-	    xerror(0, Ngt("`%ls' is not a valid integer"), countstr);
-	    SPECIAL_BI_ERROR;
+    if (iter) {
+	if (!iterinfo.iterating) {
+	    xerror(0, Ngt("not in iteration"));
 	    return Exit_ERROR;
-	} else if (countl == 0) {
-	    xerror(0, Ngt("%u: not a positive integer"), 0u);
-	    SPECIAL_BI_ERROR;
-	    return Exit_ERROR;
-	} else if (countl > UINT_MAX) {
-	    count = UINT_MAX;
-	} else {
-	    count = (unsigned) countl;
 	}
-    }
-    assert(count > 0);
-    if (execinfo.loopnest == 0) {
-	xerror(0, Ngt("not in loop"));
-	SPECIAL_BI_ERROR;
-	return Exit_ERROR;
-    }
-    if (count > execinfo.loopnest)
-	count = execinfo.loopnest;
-    if (wcscmp(ARGV(0), L"break") == 0) {
-	execinfo.breakcount = count;
+	if (wcscmp(ARGV(0), L"break") == 0) {
+	    iterinfo.exception = ie_break;
+	} else {
+	    assert(wcscmp(ARGV(0), L"continue") == 0);
+	    iterinfo.exception = ie_continue;
+	}
+	execinfo.breakcount = execinfo.loopnest;
+	return laststatus;
     } else {
-	assert(wcscmp(ARGV(0), L"continue") == 0);
-	execinfo.breakcount = count - 1;
-	execinfo.exception = ee_continue;
+	unsigned count;
+	const wchar_t *countstr = ARGV(xoptind);
+	if (countstr == NULL) {
+	    count = 1;
+	} else {
+	    unsigned long countl;
+	    if (!xwcstoul(countstr, 0, &countl)) {
+		xerror(0, Ngt("`%ls' is not a valid integer"), countstr);
+		SPECIAL_BI_ERROR;
+		return Exit_ERROR;
+	    } else if (countl == 0) {
+		xerror(0, Ngt("%u: not a positive integer"), 0u);
+		SPECIAL_BI_ERROR;
+		return Exit_ERROR;
+	    } else if (countl > UINT_MAX) {
+		count = UINT_MAX;
+	    } else {
+		count = (unsigned) countl;
+	    }
+	}
+	assert(count > 0);
+	if (execinfo.loopnest == 0) {
+	    xerror(0, Ngt("not in loop"));
+	    SPECIAL_BI_ERROR;
+	    return Exit_ERROR;
+	}
+	if (count > execinfo.loopnest)
+	    count = execinfo.loopnest;
+	if (wcscmp(ARGV(0), L"break") == 0) {
+	    execinfo.breakcount = count;
+	} else {
+	    assert(wcscmp(ARGV(0), L"continue") == 0);
+	    execinfo.breakcount = count - 1;
+	    execinfo.exception = ee_continue;
+	}
+	return Exit_SUCCESS;
     }
-    return Exit_SUCCESS;
 }
 
 const char break_help[] = Ngt(
 "break - exit loop\n"
 "\tbreak [n]\n"
-"Exits the currently executing for, while or until loop.\n"
+"\tbreak -i\n"
+"The first form exits the currently executing for/while/until loop.\n"
 "If <n> is specified, exits the <n>th outer loop.\n"
+"The second form, with the -i (--iteration) option, exits the current\n"
+"iterative execution.\n"
 );
 
 const char continue_help[] = Ngt(
 "continue - continue loop\n"
 "\tcontinue [n]\n"
-"Returns to the top of the currently executing for, while or until loop.\n"
-"If <n> is specified, returns to that of the <n>th outer loop.\n"
+"\tcontinue -i\n"
+"The first form ends the current iteration of a for/while/until loop and\n"
+"resumes the next iteration. If <n> is specified, ends the iteration of the\n"
+"<n>th outer loop.\n"
+"The second form, with the -i (--iteration) option, ends the execution of\n"
+"commands and resumes the next iteration of the current iterative execution.\n"
 );
 
 /* The "eval" builtin */
 int eval_builtin(int argc __attribute__((unused)), void **argv)
 {
+    bool iter = false;
     wchar_t opt;
     xoptind = 0, xopterr = true;
-    while ((opt = xgetopt_long(argv, L"", help_option, NULL))) {
+    while ((opt = xgetopt_long(argv, L"+i", iter_options, NULL))) {
 	switch (opt) {
+	    case L'i':
+		iter = true;
+		break;
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:
-		fprintf(stderr, gt("Usage:  eval [arg...]\n"));
+		fprintf(stderr, gt("Usage:  eval [-i] [arg...]\n"));
 		SPECIAL_BI_ERROR;
 		return Exit_ERROR;
 	}
     }
 
-    wchar_t *args = joinwcsarray(argv + xoptind, L" ");
-    exec_wcs(args, "eval", false);
-    free(args);
-    return laststatus;
+    if (iter) {
+	return exec_iteration(argv + xoptind, "eval");
+    } else {
+	wchar_t *args = joinwcsarray(argv + xoptind, L" ");
+	exec_wcs(args, "eval", false);
+	free(args);
+	return laststatus;
+    }
 }
 
 const char eval_help[] = Ngt(
 "eval - evaluate arguments as command\n"
-"\teval [arg...]\n"
-"Parses and executes the specified <arg>s as a command in the current shell\n"
-"environment.\n"
+"\teval [-i] [arg...]\n"
+"Parses and executes the specified <arg>s as commands in the current shell\n"
+"environment. Without the -i (--iteration) option, all the <arg>s are joined\n"
+"with a space inserted between each <arg> and the whole resultant string is\n"
+"parsed at a time. With the -i option, <arg>s are parsed and executed one by\n"
+"one (iterative execution).\n"
 );
 
 int dot_builtin(int argc, void **argv)
@@ -1429,8 +1587,7 @@ int dot_builtin(int argc, void **argv)
     while ((opt = xgetopt_long(argv, L"", help_option, NULL))) {
 	switch (opt) {
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:  print_usage:
 		fprintf(stderr, gt("Usage:  . file [arg...]\n"));
 		SPECIAL_BI_ERROR;
@@ -1534,8 +1691,7 @@ int exec_builtin(int argc, void **argv)
 	    case L'c':  clear = true;  break;
 	    case L'f':  force = true;  break;
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:
 		fprintf(stderr,
 		    gt("Usage:  exec [-cf] [-a name] [command [arg...]]\n"));
@@ -1691,7 +1847,7 @@ int command_builtin(int argc, void **argv)
 
     bool argv0istype = wcscmp(ARGV(0), L"type") == 0;
     bool printinfo = argv0istype, humanfriendly = argv0istype;
-    enum srchcmdtype_T type = 0;
+    bool builtin = false, external = false, defpath = false;
 
     wchar_t opt;
     xoptind = 0, xopterr = true;
@@ -1699,14 +1855,13 @@ int command_builtin(int argc, void **argv)
 		    posixly_correct ? L"+pvV" : L"+bBpvV",
 		    long_options, NULL))) {
 	switch (opt) {
-	    case L'b':  type |= sct_builtin;   break;
-	    case L'B':  type |= sct_external;  break;
-	    case L'p':  type |= sct_defpath;   break;
+	    case L'b':  builtin   = true;  break;
+	    case L'B':  external  = true;  break;
+	    case L'p':  defpath   = true;  break;
 	    case L'v':  printinfo = true;  humanfriendly = false;  break;
 	    case L'V':  printinfo = true;  humanfriendly = true;   break;
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:
 		goto print_usage;
 	}
@@ -1717,21 +1872,39 @@ int command_builtin(int argc, void **argv)
 	    goto print_usage;
 	else
 	    return Exit_SUCCESS;
-    } else if (posixly_correct && (type & (sct_builtin | sct_external))) {
+    } else if (posixly_correct && (builtin || external)) {
 	goto print_usage;
     }
 
-    if (!(type & (sct_external | sct_builtin)))
-	type |= sct_external | sct_builtin;
+    enum srchcmdtype_T type = 0;
+    if (builtin)
+	type |= sct_builtin;
+    if (external)
+	type |= sct_external;
+    if (!builtin && !external)
+	type |= sct_builtin | sct_external;
+    if (defpath)
+	type |= sct_defpath;
+
     if (!printinfo) {
 	return command_builtin_execute(
 		argc - xoptind, argv + xoptind, type);
     } else {
 	bool err = false;
-	type |= sct_function | sct_argc1;
+	bool keywordalias = false;
+	if (!builtin && !external)
+	    type |= sct_function, keywordalias = true;
+
+	clearerr(stdout);
 	for (int i = xoptind; i < argc; i++)
-	    err |= !print_command_info(ARGV(i), type, humanfriendly);
-	return err ? Exit_FAILURE : Exit_SUCCESS;
+	    err |= !print_command_info(
+		    ARGV(i), type, keywordalias, humanfriendly);
+	if (!ferror(stdout)) {
+	    return err ? Exit_FAILURE : Exit_SUCCESS;
+	} else {
+	    xerror(0, Ngt("cannot print to standard output"));
+	    return Exit_FAILURE;
+	}
     }
 
 print_usage:
@@ -1757,8 +1930,6 @@ int command_builtin_execute(int argc, void **argv, enum srchcmdtype_T type)
 
     if (!argv0)
 	argv0 = xstrdup("");
-    if (argc == 1)
-	type |= sct_argc1;
     search_command(argv0, argv[0], &ci, type);
     if (ci.type == externalprogram) {
 	pid_t cpid = fork_and_reset(0, true, t_leave);
@@ -1783,19 +1954,27 @@ int command_builtin_execute(int argc, void **argv, enum srchcmdtype_T type)
 }
 
 /* Prints info about the specified command.
- * If the command is not found, returns false. */
+ * If the command is not found, returns false.
+ * The shell keywords and aliases are printed only when `keywordalias' is true.
+ */
 bool print_command_info(
-	const wchar_t *commandname, enum srchcmdtype_T type, bool humanfriendly)
+	const wchar_t *commandname, enum srchcmdtype_T type,
+	bool keywordalias, bool humanfriendly)
 {
-#if YASH_ENABLE_ALIAS
-    if (print_alias_if_defined(commandname, humanfriendly))
-	return true;
-#endif
-    if (is_keyword(commandname)) {
+    bool found;
+
+    if (keywordalias && is_keyword(commandname)) {
 	printf(humanfriendly ? gt("%ls: shell keyword\n") : "%ls\n",
 		commandname);
 	return true;
     }
+#if YASH_ENABLE_ALIAS
+    found = keywordalias && print_alias_if_defined(commandname, humanfriendly);
+    if (found && !humanfriendly)
+	return true;
+#else
+    found = false;
+#endif
 
     char *name = malloc_wcstombs(commandname);
     if (!name)
@@ -1808,6 +1987,8 @@ bool print_command_info(
 	    if (ci.ci_path && is_executable_regular(ci.ci_path)) {
 		print_command_absolute_path(name, ci.ci_path, humanfriendly);
 		goto ok;
+	    } else if (found) {
+		goto ok;
 	    } else {
 		if (humanfriendly)
 		    xerror(0, gt("%s: no such command or function"), name);
@@ -1821,10 +2002,7 @@ bool print_command_info(
 	    goto ok;
 	case semispecialbuiltin:
 	    if (humanfriendly)
-		if (ci.ci_builtin == cd_builtin && strcmp(name, "cd") != 0)
-		    printf(gt("%s: directory\n"), name);
-		else
-		    printf(gt("%s: semi-special builtin\n"), name);
+		printf(gt("%s: semi-special builtin\n"), name);
 	    else
 		puts(name);
 	    goto ok;
@@ -1931,15 +2109,17 @@ int times_builtin(int argc __attribute__((unused)), void **argv)
     while ((opt = xgetopt_long(argv, L"", help_option, NULL))) {
 	switch (opt) {
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
-	    default:
+		return print_builtin_help(ARGV(0));
+	    default:  print_usage:
 		fprintf(stderr, gt("Usage:  times\n"));
 		SPECIAL_BI_ERROR;
 		return Exit_ERROR;
 	}
     }
 
+    if (xoptind < argc)
+	goto print_usage;
+
     double clock;
     struct tms tms;
     intmax_t sum, ssm, cum, csm;
@@ -1961,9 +2141,13 @@ int times_builtin(int argc __attribute__((unused)), void **argv)
     format_time(tms.tms_stime, ssm, sss);
     format_time(tms.tms_cutime, cum, cus);
     format_time(tms.tms_cstime, csm, css);
-    printf(Ngt("%jdm%fs %jdm%fs\n%jdm%fs %jdm%fs\n"),
-	    sum, sus, ssm, sss, cum, cus, csm, css);
-    return Exit_SUCCESS;
+    if (printf(Ngt("%jdm%fs %jdm%fs\n%jdm%fs %jdm%fs\n"),
+	    sum, sus, ssm, sss, cum, cus, csm, css) >= 0) {
+	return Exit_SUCCESS;
+    } else {
+	xerror(errno, Ngt("cannot print to standard output"));
+	return Exit_FAILURE;
+    }
 #undef format_time
 }
 
