@@ -50,7 +50,6 @@
 
 /********** Input Functions **********/
 
-static void exec_prompt_command(void);
 #if YASH_ENABLE_LINEEDIT
 static wchar_t *forward_line(wchar_t *linebuffer, xwcsbuf_T *buf)
     __attribute__((nonnull,malloc,warn_unused_result));
@@ -226,8 +225,7 @@ bool read_line_from_stdin(struct xwcsbuf_T *buf, bool trap)
 	memset(&state, 0, sizeof state);  /* initialize the state */
     }
 
-    if (!set_nonblocking(STDIN_FILENO))
-	return false;
+    set_nonblocking(STDIN_FILENO);
     while (ok) {
 	char c;
 	ssize_t n = read(STDIN_FILENO, &c, 1);
@@ -288,12 +286,18 @@ int input_readline(struct xwcsbuf_T *buf, void *inputinfo)
 
     struct parsestate_T *state = save_parse_state();
     struct input_readline_info *info = inputinfo;
+    if (info->type == 1)
+	if (!posixly_correct)
+	    exec_variable_as_commands(
+		    L VAR_PROMPT_COMMAND, VAR_PROMPT_COMMAND);
     if (do_job_control)
 	print_job_status_all(true, false, stderr);
-    if (info->type == 1) {
-	exec_prompt_command();
+    if (info->type == 1)
 	check_mail();
-    }
+    /* Note: no commands must be executed between `print_job_status_all' here
+     * and `le_readline', or the "notifyle" option won't work. More precisely,
+     * `handle_sigchld' must not be called from any other function until it is
+     * called from `wait_for_input' in the line-editing. */
 
 #if YASH_ENABLE_LINEEDIT
     /* read a line using line edit */
@@ -346,24 +350,6 @@ int input_readline(struct xwcsbuf_T *buf, void *inputinfo)
     return result;
 }
 
-/* Executes the prompt command.
- * The parse state must be saved beforehand. */
-void exec_prompt_command(void)
-{
-    if (!posixly_correct) {
-	const wchar_t *pc = getvar(L VAR_PROMPT_COMMAND);
-	if (pc) {
-	    /* PROMPT_COMMAND is executed as if a trap handler */
-	    wchar_t *savepc = xwcsdup(pc);
-	    savelaststatus = laststatus;
-	    exec_wcs(savepc, VAR_PROMPT_COMMAND, false);
-	    laststatus = savelaststatus;
-	    savelaststatus = -1;
-	    free(savepc);
-	}
-    }
-}
-
 #if YASH_ENABLE_LINEEDIT
 
 /* Processes the line buffer.
