@@ -804,8 +804,7 @@ int jobs_builtin(int argc, void **argv)
 	    case L'r':  runningonly = true;  break;
 	    case L's':  stoppedonly = true;  break;
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:
 		fprintf(stderr, gt(posixly_correct
 			    ? Ngt("Usage:  jobs [-lp] [job...]\n")
@@ -815,6 +814,8 @@ int jobs_builtin(int argc, void **argv)
     }
 
     nextforceexit = true;
+
+    clearerr(stdout);
     if (xoptind < argc) {
 	do {
 	    const wchar_t *jobspec = ARGV(xoptind);
@@ -846,7 +847,12 @@ int jobs_builtin(int argc, void **argv)
 	}
     }
 
-    return err ? Exit_FAILURE : Exit_SUCCESS;
+    if (!ferror(stdout)) {
+	return err ? Exit_FAILURE : Exit_SUCCESS;
+    } else {
+	xerror(0, Ngt("cannot print to standard output"));
+	return Exit_FAILURE;
+    }
 }
 
 /* Prints the job status */
@@ -903,9 +909,8 @@ int fg_builtin(int argc, void **argv)
     while ((opt = xgetopt_long(argv, L"", help_option, NULL))) {
 	switch (opt) {
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
-	    default:
+		return print_builtin_help(ARGV(0));
+	    default:  print_usage:
 		fprintf(stderr, gt(fg ? Ngt("Usage:  fg [job]\n")
 		                      : Ngt("Usage:  bg [job...]\n")));
 		return Exit_ERROR;
@@ -920,8 +925,11 @@ int fg_builtin(int argc, void **argv)
     int status = Exit_SUCCESS;
     job_T *job;
 
-    nextforceexit = true;
     if (xoptind < argc) {
+	if (fg && posixly_correct && argc - xoptind > 1) {
+	    xerror(0, Ngt("too many operands"));
+	    goto print_usage;
+	}
 	do {
 	    const wchar_t *jobspec = ARGV(xoptind);
 	    if (jobspec[0] == L'%') {
@@ -979,19 +987,24 @@ int continue_job(size_t jobnumber, job_T *job, bool fg)
     if (job->j_status != JS_DONE) {
 	if (fg)
 	    put_foreground(job->j_pgid);
-	if (job->j_status == JS_STOPPED || !posixly_correct)
-	    if (kill(-job->j_pgid, SIGCONT) < 0)
-		xerror(errno, "kill(%jd,CONT)", (intmax_t) -job->j_pgid);
-	job->j_status = JS_RUNNING;
+	if (kill(-job->j_pgid, SIGCONT) >= 0)
+	    job->j_status = JS_RUNNING;
+    } else {
+	if (!fg)
+	    xerror(0, Ngt("job %%%zu has already terminated"), jobnumber);
     }
 
-    if (fg)
+    int status;
+    if (fg) {
 	wait_for_job(jobnumber, true, false, false);
-    int status = (job->j_status == JS_RUNNING)
-	? Exit_SUCCESS : calc_status_of_job(job);
-    if (job->j_status == JS_DONE) {
-	notify_signaled_job(jobnumber);
-	remove_job(jobnumber);
+	status = (job->j_status == JS_RUNNING)
+	    ? Exit_SUCCESS : calc_status_of_job(job);
+	if (job->j_status == JS_DONE) {
+	    notify_signaled_job(jobnumber);
+	    remove_job(jobnumber);
+	}
+    } else {
+	status = (job->j_status == JS_RUNNING) ? Exit_SUCCESS : Exit_FAILURE;
     }
     return status;
 }
@@ -1024,16 +1037,13 @@ int wait_builtin(int argc, void **argv)
     while ((opt = xgetopt_long(argv, L"", help_option, NULL))) {
 	switch (opt) {
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:
 		fprintf(stderr, gt("Usage:  wait [job or pid...]\n"));
 		return Exit_ERROR;
 	}
     }
 
-    nextforceexit = true;
-
     job_T *job;
     if (xoptind < argc) {
 	do {
@@ -1127,15 +1137,13 @@ int disown_builtin(int argc, void **argv)
 	switch (opt) {
 	    case L'a':  all = true;  break;
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:
 		fprintf(stderr, gt("Usage:  disown [job...]\n"));
 		return Exit_ERROR;
 	}
     }
 
-    nextforceexit = true;
     if (all) {
 	remove_all_jobs();
     } else if (xoptind < argc) {
