@@ -19,15 +19,12 @@
 #include "../common.h"
 #include <assert.h>
 #include <errno.h>
-#include <stdarg.h>
 #include <stdbool.h>
-#include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 #include <wchar.h>
-#include <wctype.h>
 #include "../option.h"
 #include "../sig.h"
 #include "../strbuf.h"
@@ -50,8 +47,6 @@ static inline bool has_meta_bit(char c)
 static inline trieget_T make_trieget(const wchar_t *keyseq)
     __attribute__((nonnull,const));
 static void append_to_second_buffer(wchar_t wc);
-static inline wchar_t wb_get_char(const xwcsbuf_T *buf)
-    __attribute__((nonnull,pure));
 
 
 /* The state of lineedit. */
@@ -66,12 +61,7 @@ enum le_editstate_T le_editstate;
  * afterward. */
 bool le_setup(void)
 {
-    static bool initialized = false;
-    if (!initialized) {
-	initialized = true;
-	le_keymap_init();
-    }
-
+    le_keymap_init();
     return isatty(STDIN_FILENO) && isatty(STDERR_FILENO)
 	&& le_setupterm(true)
 	&& le_set_terminal();
@@ -93,6 +83,7 @@ wchar_t *le_readline(const wchar_t *prompt)
     le_state = LE_STATE_ACTIVE;
     le_editing_init();
     le_display_init(prompt);
+    le_display_maybe_promptsp();
     reader_init();
     le_editstate = LE_EDITSTATE_EDITING;
 
@@ -141,8 +132,8 @@ void le_resume_readline(void)
 	le_state = LE_STATE_ACTIVE;
 	le_setupterm(true);
 	le_set_terminal();
-	le_display_print_all(true);
-	le_display_reposition_cursor();
+	le_display_maybe_promptsp();
+	le_display_update();
 	fflush(stderr);
     }
 }
@@ -154,8 +145,7 @@ void le_display_size_changed(void)
     if (le_state == LE_STATE_ACTIVE) {
 	le_display_clear();
 	le_setupterm(true);
-	le_display_print_all(false);
-	le_display_reposition_cursor();
+	le_display_update();
 	fflush(stderr);
     }
 }
@@ -212,6 +202,7 @@ void read_next(void)
 	goto direct_first_buffer;
 
     /* wait for and read the next byte */
+    le_display_update();
     fflush(stderr);
     timeout = !wait_for_input(
 	    STDIN_FILENO, true, keycode_ambiguous ? le_read_timeout : -1);
@@ -319,22 +310,35 @@ process_wide:
     /* process key mapping for wide characters */
 process_keymap:
     while (reader_second_buffer.length > 0) {
+	register wchar_t c;
 	trieget_T tg = trie_getw(
 		le_current_mode->keymap, reader_second_buffer.contents);
 	switch (tg.type) {
 	    case TG_NOMATCH:
-		le_invoke_command(le_current_mode->default_command,
-			wb_get_char(&reader_second_buffer));
+		assert(reader_second_buffer.length > 0);
+		if (reader_second_buffer.length > 1)
+		    c = L'\0';
+		else
+		    c = reader_second_buffer.contents[0];
+		le_invoke_command(le_current_mode->default_command, c);
 		wb_clear(&reader_second_buffer);
 		break;
 	    case TG_UNIQUE:
-		le_invoke_command(tg.value.cmdfunc,
-			wb_get_char(&reader_second_buffer));
+		assert(tg.matchlength > 0);
+		c = reader_second_buffer.contents[tg.matchlength - 1];
+		le_invoke_command(tg.value.cmdfunc, c);
 		wb_remove(&reader_second_buffer, 0, tg.matchlength);
 		break;
 	    case TG_NEEDMORE:
 	    case TG_AMBIGUOUS:
 		return;
+	/* For an unmatched control sequence, `cmd_self_insert' should not
+	 * receive any part of the sequence as argument (because the sequence
+	 * should not be inserted in the main buffer), so the input is passed
+	 * to the default command iff the input is a usual character.
+	 * For a matched control sequence, the last character of the sequence
+	 * is passed to the command so that commands like `cmd_digit_argument'
+	 * work. */
 	}
 	if (le_editstate != LE_EDITSTATE_EDITING)
 	    break;
@@ -401,11 +405,5 @@ void append_to_second_buffer(wchar_t wc)
     le_next_verbatim = false;
 }
 
-wchar_t wb_get_char(const xwcsbuf_T *buf)
-{
-    assert(buf->length > 0);
-    return buf->length > 1 ? L'\0' : buf->contents[0];
-}
-
 
 /* vim: set ts=8 sts=4 sw=4 noet tw=80: */
