@@ -43,6 +43,7 @@
 /* terminfo capabilities */
 #define TI_am      "am"
 #define TI_bel     "bel"
+#define TI_clear   "clear"
 #define TI_cols    "cols"
 #define TI_cr      "cr"
 #define TI_cub     "cub"
@@ -322,7 +323,7 @@ _Bool le_setupterm(_Bool bypass)
     return 1;
 }
 
-/* Initialized `le_keycodes'. */
+/* Initializes `le_keycodes'. */
 void set_up_keycodes(void)
 {
     trie_destroy(le_keycodes);
@@ -624,6 +625,17 @@ _Bool le_print_ed(void)
 	return 0;
 }
 
+/* Prints "clear" variable if available. (clear whole screen)
+ * Returns true iff successful. */
+_Bool le_print_clear(void)
+{
+    char *v = tigetstr(TI_clear);
+    if (is_strcap_valid(v))
+	return tputs(v, le_lines, putchar_stderr) != ERR;
+    else
+	return 0;
+}
+
 /* Prints "sgr" variable. Every argument must be 0 or 1. */
 void le_print_sgr(long standout, long underline, long reverse, long blink,
 	long dim, long bold, long invisible)
