@@ -18,6 +18,7 @@
 
 #include "common.h"
 #include <assert.h>
+#include <errno.h>
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -116,9 +117,6 @@ bool shopt_xtrace;
 bool shopt_curasync;
 /* If set, lines that start with a space are not saved in the history. */
 bool shopt_histspace;
-/* If set, the shell changes the working directory when a command is a
- * directory rather than a executable. */
-bool shopt_autocd;
 
 /* If set, don't perform filename expansions.
  * Corresponds to -f/--noglob option. */
@@ -152,7 +150,7 @@ typedef enum shopt_index_T {
     SHOPT_RCFILE, SHOPT_VERSION,
     SHOPT_ALLEXPORT, SHOPT_HASHONDEF, SHOPT_NOCLOBBER, SHOPT_NOGLOB,
     SHOPT_NOCASEGLOB, SHOPT_DOTGLOB, SHOPT_MARKDIRS, SHOPT_EXTENDEDGLOB,
-    SHOPT_NULLGLOB, SHOPT_BRACEEXPAND, SHOPT_CURASYNC, SHOPT_AUTOCD,
+    SHOPT_NULLGLOB, SHOPT_BRACEEXPAND, SHOPT_CURASYNC,
     SHOPT_ERREXIT, SHOPT_NOUNSET, SHOPT_NOEXEC, SHOPT_IGNOREEOF, SHOPT_VERBOSE,
     SHOPT_XTRACE, SHOPT_HISTSPACE, SHOPT_NOLOG, SHOPT_MONITOR, SHOPT_NOTIFY,
 #if YASH_ENABLE_LINEEDIT
@@ -160,7 +158,7 @@ typedef enum shopt_index_T {
 #endif
     SHOPT_POSIX,
 #if YASH_ENABLE_LINEEDIT
-    SHOPT_VI, /* SHOPT_EMACS, */
+    SHOPT_VI, SHOPT_EMACS,
     SHOPT_LE_CONVMETA, SHOPT_LE_NOCONVMETA, SHOPT_LE_PROMPTSP,
 #endif
     SHOPT_HELP,
@@ -188,7 +186,6 @@ static const struct xoption long_options[] = {
     [SHOPT_NULLGLOB]      = { L"nullglob",     xno_argument, L'L', },
     [SHOPT_BRACEEXPAND]   = { L"braceexpand",  xno_argument, L'L', },
     [SHOPT_CURASYNC]      = { L"curasync",     xno_argument, L'L', },
-    [SHOPT_AUTOCD]        = { L"autocd",       xno_argument, L'L', },
     [SHOPT_ERREXIT]       = { L"errexit",      xno_argument, L'L', },
     [SHOPT_NOUNSET]       = { L"nounset",      xno_argument, L'L', },
     [SHOPT_NOEXEC]        = { L"noexec",       xno_argument, L'L', },
@@ -205,7 +202,7 @@ static const struct xoption long_options[] = {
     [SHOPT_POSIX]         = { L"posix",        xno_argument, L'L', },
 #if YASH_ENABLE_LINEEDIT
     [SHOPT_VI]            = { L"vi",           xno_argument, L'L', },
-    // [SHOPT_EMACS]         = { L"emacs",        xno_argument, L'L', },
+    [SHOPT_EMACS]         = { L"emacs",        xno_argument, L'L', },
     [SHOPT_LE_CONVMETA]   = { L"le-convmeta",  xno_argument, L'L', },
     [SHOPT_LE_NOCONVMETA] = { L"le-noconvmeta",xno_argument, L'L', },
     [SHOPT_LE_PROMPTSP]   = { L"le-promptsp",  xno_argument, L'L', },
@@ -227,7 +224,6 @@ static const struct setoptinfo_T setoptinfo[] = {
     [SHOPT_NULLGLOB]      = { set_bool_option, &shopt_nullglob, },
     [SHOPT_BRACEEXPAND]   = { set_bool_option, &shopt_braceexpand, },
     [SHOPT_CURASYNC]      = { set_bool_option, &shopt_curasync, },
-    [SHOPT_AUTOCD]        = { set_bool_option, &shopt_autocd, },
     [SHOPT_ERREXIT]       = { set_bool_option, &shopt_errexit, },
     [SHOPT_NOUNSET]       = { set_bool_option, &shopt_nounset, },
     [SHOPT_NOEXEC]        = { set_bool_option, &shopt_noexec, },
@@ -244,7 +240,7 @@ static const struct setoptinfo_T setoptinfo[] = {
     [SHOPT_POSIX]         = { set_bool_option, &posixly_correct, },
 #if YASH_ENABLE_LINEEDIT
     [SHOPT_VI]            = { set_lineedit_option, NULL, },
-    //[SHOPT_EMACS]         = { set_lineedit_option, NULL, },
+    [SHOPT_EMACS]         = { set_lineedit_option, NULL, },
     [SHOPT_LE_CONVMETA]   = { set_le_convmeta_option, NULL, },
     [SHOPT_LE_NOCONVMETA] = { set_le_noconvmeta_option, NULL, },
     [SHOPT_LE_PROMPTSP]   = { set_bool_option, &shopt_le_promptsp, },
@@ -351,7 +347,7 @@ void set_lineedit_option(void *argp __attribute__((unused)))
 
     switch (setoptindex) {
 	case SHOPT_VI:     opt = shopt_vi;     break;
-	//case SHOPT_EMACS:  opt = shopt_emacs;  break;
+	case SHOPT_EMACS:  opt = shopt_emacs;  break;
 	default:           return;
     }
     if (xoptopt == L'-') {
@@ -416,8 +412,8 @@ wchar_t *get_hyphen_parameter(void)
 
 /********** Builtin **********/
 
-static void set_builtin_print_current_settings(void);
-static void set_builtin_print_restoring_commands(void);
+static int set_builtin_print_current_settings(void);
+static int set_builtin_print_restoring_commands(void);
 
 int set_builtin(int argc, void **argv)
 {
@@ -428,11 +424,9 @@ int set_builtin(int argc, void **argv)
     }
     if (argc == 2) {
 	if (wcscmp(ARGV(1), L"-o") == 0) {
-	    set_builtin_print_current_settings();
-	    return Exit_SUCCESS;
+	    return set_builtin_print_current_settings();
 	} else if (wcscmp(ARGV(1), L"+o") == 0) {
-	    set_builtin_print_restoring_commands();
-	    return Exit_SUCCESS;
+	    return set_builtin_print_restoring_commands();
 	}
     }
 
@@ -455,8 +449,7 @@ int set_builtin(int argc, void **argv)
 		}
 		break;
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    case L'?':  optionerror:
 		fprintf(stderr,
 		    gt("Usage:  set [-abefhmnuvxC] [-o option] [+o option] "
@@ -477,22 +470,22 @@ int set_builtin(int argc, void **argv)
     return Exit_SUCCESS;
 }
 
-void set_builtin_print_current_settings(void)
+int set_builtin_print_current_settings(void)
 {
+    bool err = false;
     const char *vals[] = {
 	[true]  = gt("yes"),
 	[false] = gt("no"),
     };
 #define PRINTSETTING(name,value) \
-    printf("%-15ls %s\n", L"" #name, vals[(bool) (value)])
+    (err |= printf("%-15ls %s\n", L"" #name, vals[(bool) (value)]) < 0)
 
     PRINTSETTING(allexport, shopt_allexport);
-    PRINTSETTING(autocd, shopt_autocd);
     PRINTSETTING(braceexpand, shopt_braceexpand);
     PRINTSETTING(curasync, shopt_curasync);
     PRINTSETTING(dotglob, shopt_dotglob);
 #if YASH_ENABLE_LINEEDIT
-    //PRINTSETTING(emacs, shopt_lineedit == shopt_emacs);
+    PRINTSETTING(emacs, shopt_lineedit == shopt_emacs);
 #endif
     PRINTSETTING(errexit, shopt_errexit);
     PRINTSETTING(extendedglob, shopt_extendedglob);
@@ -526,20 +519,27 @@ void set_builtin_print_current_settings(void)
 #endif
     PRINTSETTING(xtrace, shopt_xtrace);
 #undef PRINTSETTING
+
+    if (!err) {
+	return Exit_SUCCESS;
+    } else {
+	xerror(errno, Ngt("cannot print to standard output"));
+	return Exit_FAILURE;
+    }
 }
 
-void set_builtin_print_restoring_commands(void)
+int set_builtin_print_restoring_commands(void)
 {
+    bool err = false;
 #define PRINTSETTING(name,value) \
-    printf("set %co %ls\n", (value) ? '-' : '+', L"" #name)
+    (err |= printf("set %co %ls\n", (value) ? '-' : '+', L"" #name) < 0)
 
     PRINTSETTING(allexport, shopt_allexport);
-    PRINTSETTING(autocd, shopt_autocd);
     PRINTSETTING(braceexpand, shopt_braceexpand);
     PRINTSETTING(curasync, shopt_curasync);
     PRINTSETTING(dotglob, shopt_dotglob);
 #if YASH_ENABLE_LINEEDIT
-    //PRINTSETTING(emacs, shopt_lineedit == shopt_emacs);
+    PRINTSETTING(emacs, shopt_lineedit == shopt_emacs);
 #endif
     PRINTSETTING(errexit, shopt_errexit);
     PRINTSETTING(extendedglob, shopt_extendedglob);
@@ -573,6 +573,13 @@ void set_builtin_print_restoring_commands(void)
 #endif
     PRINTSETTING(xtrace, shopt_xtrace);
 #undef PRINTSETTING
+
+    if (!err) {
+	return Exit_SUCCESS;
+    } else {
+	xerror(errno, Ngt("cannot print to standard output"));
+	return Exit_FAILURE;
+    }
 }
 
 const char set_help[] = Ngt(
@@ -647,15 +654,12 @@ const char set_help[] = Ngt(
 "\tWhen a new background job is invoked, it becomes the current job.\n"
 " --histspace\n"
 "\tDo not save lines that start with a space in the history.\n"
-" --autocd\n"
-"\tIf a simple command cannot be executed but it is a directory name,\n"
-"\tthe shell performs the `cd' command to that directory.\n"
 " --posix\n"
 "\tMake the shell behave as the POSIX shell.\n"
 " --vi\n"
 "\tEnable vi-like editing.\n"
-//" --emacs\n"
-//"\tEnable emacs-like editing.\n"
+" --emacs\n"
+"\tEnable emacs-like editing.\n"
 " --le-convmeta\n"
 "\tTreat 8th bit of input as a meta-key flag (regardless of terminfo).\n"
 " --le-noconvmeta\n"
