@@ -1019,10 +1019,14 @@ bool is_reentry(const struct stat *st, const plist_T *dirstack)
 static int change_directory(
 	const wchar_t *newpwd, bool printnewdir, bool logical)
     __attribute__((nonnull,warn_unused_result));
-static void print_command_paths(bool all);
-static void print_home_directories(void);
-static void print_umask_octal(mode_t mode);
-static void print_umask_symbolic(mode_t mode);
+static void canonicalize_path_ex(xwcsbuf_T *buf)
+    __attribute__((nonnull));
+static bool starts_with_root_parent(const wchar_t *path)
+    __attribute__((nonnull,pure));
+static bool print_command_paths(bool all);
+static bool print_home_directories(void);
+static bool print_umask_octal(mode_t mode);
+static bool print_umask_symbolic(mode_t mode);
 static int set_umask(const wchar_t *newmask)
     __attribute__((nonnull));
 static inline mode_t copy_user_mask(mode_t mode)
@@ -1044,8 +1048,6 @@ static const struct xoption cd_pwd_options[] = {
  * -L: don't resolve symlinks (default)
  * -P: resolve symlinks
  * -L and -P are mutually exclusive: the one specified last is used. */
-/* When the `autocd' option is enabled, this builtin may be called with an
- * argument which is a directory name rather than the string "cd". */
 int cd_builtin(int argc, void **argv)
 {
     bool logical = true;
@@ -1054,18 +1056,15 @@ int cd_builtin(int argc, void **argv)
     wchar_t opt;
 
     xoptind = 0, xopterr = true;
-    if (wcscmp(ARGV(0), L"cd") == 0) {
-	while ((opt = xgetopt_long(argv, L"-LP", cd_pwd_options, NULL))) {
-	    switch (opt) {
-		case L'L':  logical = true;   break;
-		case L'P':  logical = false;  break;
-		case L'-':
-		    print_builtin_help(ARGV(0));
-		    return Exit_SUCCESS;
-		default:
-		    fprintf(stderr, gt("Usage:  %ls [-L|-P] [dir]\n"), L"cd");
-		    return Exit_ERROR;
-	    }
+    while ((opt = xgetopt_long(argv, L"-LP", cd_pwd_options, NULL))) {
+	switch (opt) {
+	    case L'L':  logical = true;   break;
+	    case L'P':  logical = false;  break;
+	    case L'-':
+		return print_builtin_help(ARGV(0));
+	    default:
+		fprintf(stderr, gt("Usage:  %ls [-L|-P] [dir]\n"), L"cd");
+		return Exit_ERROR;
 	}
     }
 
@@ -1099,7 +1098,6 @@ int change_directory(const wchar_t *newpwd, bool printnewdir, bool logical)
     const wchar_t *oldpwd;
     xwcsbuf_T curpath;
     size_t curpathoffset = 0;
-    bool err = false;
 
     /* get the current value of $PWD as `oldpwd' */
     oldpwd = getvar(L VAR_PWD);
@@ -1113,8 +1111,10 @@ int change_directory(const wchar_t *newpwd, bool printnewdir, bool logical)
 
 	char *pwd = xgetcwd();
 	if (pwd == NULL) {
-	    xerror(errno, Ngt("cannot determine current directory"));
-	    err = true;
+	    if (logical) {
+		xerror(errno, Ngt("cannot determine current directory"));
+		return Exit_FAILURE;
+	    }
 	} else {
 	    wchar_t *wpwd = realloc_mbstowcs(pwd);
 	    if (wpwd != NULL) {
@@ -1201,7 +1201,7 @@ step7:
 
     /* step 9: determine `curpathoffset' */
     assert(logical);
-    {
+    if (oldpwd[wcsspn(oldpwd, L"/")] != L'\0') {
 	wchar_t *s = matchwcsprefix(curpath.contents, oldpwd);
 	if (s != NULL && *s == L'/') {
 	    s++;
@@ -1229,42 +1229,57 @@ step10:  /* do chdir */
     }
 
 #ifndef NDEBUG
-    newpwd = NULL;  /* `newpwd' must not be used any more */
+    newpwd = NULL;
+    /* `newpwd' must not be used any more because it may be pointing to the
+     * current value of $OLDPWD, which is going to be re-assigned to. */
 #endif
 
     /* set $OLDPWD and $PWD */
     if (oldpwd != NULL)
-	if (!set_variable(L VAR_OLDPWD, xwcsdup(oldpwd), SCOPE_GLOBAL, false))
-	    err = true;
+	set_variable(L VAR_OLDPWD, xwcsdup(oldpwd), SCOPE_GLOBAL, false);
     if (logical) {
+	if (!posixly_correct)
+	    canonicalize_path_ex(&curpath);
 	if (printnewdir)
 	    printf("%ls\n", curpath.contents);
-	if (!set_variable(L VAR_PWD, wb_towcs(&curpath), SCOPE_GLOBAL, false))
-	    err = true;
+	set_variable(L VAR_PWD, wb_towcs(&curpath), SCOPE_GLOBAL, false);
     } else {
 	wb_destroy(&curpath);
 
 	char *mbsnewpwd = xgetcwd();
-	if (mbsnewpwd == NULL) {
-	    xerror(errno, Ngt("cannot determine new working directory"));
-	    err = true;
-	} else {
+	if (mbsnewpwd != NULL) {
 	    if (printnewdir)
 		printf("%s\n", mbsnewpwd);
 
 	    wchar_t *wnewpwd = realloc_mbstowcs(mbsnewpwd);
-	    if (wnewpwd == NULL) {
-		xerror(EILSEQ, Ngt("cannot convert multibyte characters "
-			    "into wide characters"));
-		err = true;
-	    } else {
-		if (!set_variable(L VAR_PWD, wnewpwd, SCOPE_GLOBAL, false))
-		    err = true;
-	    }
+	    if (wnewpwd != NULL)
+		set_variable(L VAR_PWD, wnewpwd, SCOPE_GLOBAL, false);
 	}
     }
+    if (!posixly_correct)
+	exec_variable_as_commands(L VAR_YASH_AFTER_CD, VAR_YASH_AFTER_CD);
 
-    return err ? Exit_FAILURE : Exit_SUCCESS;
+    return Exit_SUCCESS;
+}
+
+/* Removes "/.." components at the beginning of the string in the buffer
+ * if the root directory and its parent are the same directory. */
+void canonicalize_path_ex(xwcsbuf_T *buf)
+{
+    if (starts_with_root_parent(buf->contents) && is_same_file("/", "/..")) {
+	do {
+	    wb_remove(buf, 0, 3);
+	} while (starts_with_root_parent(buf->contents));
+	if (buf->length == 0)
+	    wb_wccat(buf, L'/');
+    }
+}
+
+/* Returns true iff the given path starts with "/..". */
+bool starts_with_root_parent(const wchar_t *path)
+{
+    return path[0] == L'/' && path[1] == L'.' && path[2] == L'.' &&
+	(path[3] == L'\0' || path[3] == L'/');
 }
 
 const char cd_help[] = Ngt(
@@ -1300,8 +1315,7 @@ int pushd_builtin(int argc __attribute__((unused)), void **argv)
 	    case L'L':  logical = true;   break;
 	    case L'P':  logical = false;  break;
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:
 		fprintf(stderr, gt("Usage:  %ls [-L|-P] [dir]\n"), L"pushd");
 		return Exit_ERROR;
@@ -1373,8 +1387,7 @@ int popd_builtin(int argc __attribute__((unused)), void **argv)
     while ((opt = xgetopt_long(argv, L"-", help_option, NULL))) {
 	switch (opt) {
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:
 		fprintf(stderr, gt("Usage:  popd [index]\n"));
 		return Exit_ERROR;
@@ -1396,7 +1409,7 @@ int popd_builtin(int argc __attribute__((unused)), void **argv)
     if (!parse_dirstack_index(arg, &stackindex, &dummy, true))
 	return Exit_FAILURE;
     if (stackindex == SIZE_MAX) {
-	xerror(0, Ngt("`%ls' is not a valid integer"), arg);
+	xerror(0, Ngt("`%ls' is not a valid index"), arg);
 	return Exit_ERROR;
     }
 
@@ -1437,6 +1450,7 @@ int pwd_builtin(int argc __attribute__((unused)), void **argv)
 {
     bool logical = true;
     char *mbspwd;
+    int result;
     wchar_t opt;
 
     xoptind = 0, xopterr = true;
@@ -1445,24 +1459,23 @@ int pwd_builtin(int argc __attribute__((unused)), void **argv)
 	    case L'L':  logical = true;   break;
 	    case L'P':  logical = false;  break;
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
-	    default:
+		return print_builtin_help(ARGV(0));
+	    default:  print_usage:
 		fprintf(stderr, gt("Usage:  pwd [-L|-P]\n"));
 		return Exit_ERROR;
 	}
     }
 
+    if (xoptind != argc)
+	goto print_usage;
+
     if (logical) {
 	const wchar_t *pwd = getvar(L VAR_PWD);
 	if (pwd != NULL && pwd[0] == L'/' && is_normalized_path(pwd)) {
 	    mbspwd = malloc_wcstombs(pwd);
 	    if (mbspwd != NULL) {
-		if (is_same_file(mbspwd, ".")) {
-		    printf("%s\n", mbspwd);
-		    free(mbspwd);
-		    return Exit_SUCCESS;
-		}
+		if (is_same_file(mbspwd, "."))
+		    goto print;
 		free(mbspwd);
 	    }
 	}
@@ -1473,9 +1486,15 @@ int pwd_builtin(int argc __attribute__((unused)), void **argv)
 	xerror(errno, Ngt("cannot determine current directory"));
 	return Exit_FAILURE;
     }
-    printf("%s\n", mbspwd);
+print:
+    if (printf("%s\n", mbspwd) >= 0) {
+	result = Exit_SUCCESS;
+    } else {
+	xerror(errno, Ngt("cannot print to standard output"));
+	result = Exit_FAILURE;
+    }
     free(mbspwd);
-    return Exit_SUCCESS;
+    return result;
 }
 
 const char pwd_help[] = Ngt(
@@ -1515,8 +1534,7 @@ int hash_builtin(int argc, void **argv)
 	    case L'd':  dir    = true;  break;
 	    case L'r':  remove = true;  break;
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:
 		goto print_usage;
 	}
@@ -1535,7 +1553,7 @@ int hash_builtin(int argc, void **argv)
 	    }
 	} else {
 	    if (xoptind == argc) {  // print all
-		print_home_directories();
+		err = print_home_directories();
 	    } else {                // remember the specified
 		for (int i = xoptind; i < argc; i++) {
 		    if (!get_home_directory(ARGV(i), true)) {
@@ -1555,14 +1573,12 @@ int hash_builtin(int argc, void **argv)
 		    if (cmd) {
 			forget_command_path(cmd);
 			free(cmd);
-		    } else {
-			err |= true;
 		    }
 		}
 	    }
 	} else {
 	    if (xoptind == argc) {  // print all
-		print_command_paths(all);
+		err = print_command_paths(all);
 	    } else {                // remember the specified
 		for (int i = xoptind; i < argc; i++) {
 		    if (wcschr(ARGV(i), L'/')) {
@@ -1579,8 +1595,6 @@ int hash_builtin(int argc, void **argv)
 			    err = true;
 			}
 			free(cmd);
-		    } else {
-			err |= true;
 		    }
 		}
 	    }
@@ -1596,26 +1610,39 @@ print_usage:
     return Exit_ERROR;
 }
 
-/* Prints the entries of the command hashtable. */
-void print_command_paths(bool all)
+/* Prints the entries of the command hashtable.
+ * Returns FALSE iff successful. */
+bool print_command_paths(bool all)
 {
     kvpair_T kv;
     size_t index = 0;
 
-    while ((kv = ht_next(&cmdhash, &index)).key)
-	if (all || !get_builtin(kv.key))
-	    puts(kv.value);
+    while ((kv = ht_next(&cmdhash, &index)).key) {
+	if (all || !get_builtin(kv.key)) {
+	    if (printf("%s\n", (char *) kv.value) < 0) {
+		xerror(errno, Ngt("cannot print to standard output"));
+		return true;
+	    }
+	}
+    }
+    return false;
 }
 
-/* Prints the entries of the home directory hashtable. */
-void print_home_directories(void)
+/* Prints the entries of the home directory hashtable.
+ * Returns FALSE iff successful. */
+bool print_home_directories(void)
 {
     kvpair_T kv;
     size_t index = 0;
 
-    while ((kv = ht_next(&homedirhash, &index)).key)
-	printf("~%ls=%ls\n",
-		(const wchar_t *) kv.key, (const wchar_t *) kv.value);
+    while ((kv = ht_next(&homedirhash, &index)).key) {
+	const wchar_t *key = kv.key, *value = kv.value;
+	if (printf("~%ls=%ls\n", key, value) < 0) {
+	    xerror(errno, Ngt("cannot print to standard output"));
+	    return true;
+	}
+    }
+    return false;
 }
 
 const char hash_help[] = Ngt(
@@ -1658,8 +1685,7 @@ int umask_builtin(int argc, void **argv)
 		symbolic = true;
 		break;
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:
 		goto print_usage;
 	}
@@ -1669,28 +1695,34 @@ int umask_builtin(int argc, void **argv)
 	mode_t mode = umask(0);
 	umask(mode);
 	if (symbolic)
-	    print_umask_symbolic(mode);
+	    return print_umask_symbolic(mode);
 	else
-	    print_umask_octal(mode);
+	    return print_umask_octal(mode);
     } else if (xoptind + 1 == argc) {
 	return set_umask(ARGV(xoptind));
     } else {
 	goto print_usage;
     }
-    return Exit_SUCCESS;
 
 print_usage:
     fprintf(stderr, gt("Usage:  umask [-S] [mask]\n"));
     return Exit_ERROR;
 }
 
-void print_umask_octal(mode_t mode)
+bool print_umask_octal(mode_t mode)
 {
-    printf("0%.3jo\n", (uintmax_t) mode);
+    if (printf("0%.3jo\n", (uintmax_t) mode) >= 0) {
+	return Exit_SUCCESS;
+    } else {
+	xerror(errno, Ngt("cannot print to standard output"));
+	return Exit_FAILURE;
+    }
 }
 
-void print_umask_symbolic(mode_t mode)
+bool print_umask_symbolic(mode_t mode)
 {
+    clearerr(stdout);
+
     putchar('u');
     putchar('=');
     if (!(mode & S_IRUSR)) putchar('r');
@@ -1709,6 +1741,13 @@ void print_umask_symbolic(mode_t mode)
     if (!(mode & S_IWOTH)) putchar('w');
     if (!(mode & S_IXOTH)) putchar('x');
     putchar('\n');
+
+    if (!ferror(stdout)) {
+	return Exit_SUCCESS;
+    } else {
+	xerror(errno, Ngt("cannot print to standard output"));
+	return Exit_FAILURE;
+    }
 }
 
 int set_umask(const wchar_t *maskstr)
