@@ -625,9 +625,13 @@ void handle_sigchld(void)
         do_wait();
 #if YASH_ENABLE_LINEEDIT
 	if (shopt_notify || (shopt_notifyle && le_state == LE_STATE_ACTIVE)) {
-	    le_suspend_readline();
-	    print_job_status_all(true, false, stderr);
-	    le_resume_readline();
+	    if (le_state == LE_STATE_ACTIVE) {
+		le_suspend_readline();
+		print_job_status_all(true, false, stderr);
+		le_resume_readline();
+	    } else {
+		print_job_status_all(true, false, stderr);
+	    }
 	}
 #else
 	if (shopt_notify) {
@@ -959,7 +963,7 @@ void reset_sigwinch(void)
 
 /********** Builtin **********/
 
-static void print_trap(const char *signame, const wchar_t *command)
+static bool print_trap(const char *signame, const wchar_t *command)
     __attribute__((nonnull(1)));
 static bool print_signal(int signum, const char *name, bool verbose);
 static bool signal_job(int signum, const wchar_t *jobname)
@@ -986,8 +990,7 @@ int trap_builtin(int argc, void **argv)
 		print = true;
 		break;
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:
 		goto print_usage;
 	}
@@ -1001,7 +1004,8 @@ int trap_builtin(int argc, void **argv)
 	for (const signal_T *s = signals; s->no; s++) {
 	    if (!sigismember(&ss, s->no)) {
 		sigaddset(&ss, s->no);
-		print_trap(s->name, trap_command[sigindex(s->no)]);
+		if (!print_trap(s->name, trap_command[sigindex(s->no)]))
+		    return Exit_FAILURE;
 	    }
 	}
 #if defined SIGRTMIN && defined SIGRTMAX
@@ -1009,12 +1013,14 @@ int trap_builtin(int argc, void **argv)
 	for (int i = 0; i < RTSIZE; i++) {
 	    if (sigrtmin + i > sigrtmax)
 		break;
-	    print_trap(get_signal_name(sigrtmin + i), rttrap_command[i]);
+	    if (!print_trap(get_signal_name(sigrtmin + i), rttrap_command[i]))
+		return Exit_FAILURE;
 	}
 #endif
 	return Exit_SUCCESS;
     } else if (print) {
 	/* print specified traps */
+	bool ok = true;
 #if defined SIGRTMIN && defined SIGRTMAX
 	int sigrtmin = SIGRTMIN, sigrtmax = SIGRTMAX;
 #endif
@@ -1033,17 +1039,17 @@ int trap_builtin(int argc, void **argv)
 		if (sigrtmin <= signum && signum <= sigrtmax) {
 		    int index = signum - sigrtmin;
 		    if (index < RTSIZE)
-			print_trap(get_signal_name(signum),
+			ok = print_trap(get_signal_name(signum),
 				rttrap_command[index]);
 		} else
 #endif
 		{
-		    print_trap(name, trap_command[sigindex(signum)]);
+		    ok = print_trap(name, trap_command[sigindex(signum)]);
 		}
 	    }
 	    free(name);
-	} while (++xoptind < argc);
-	return Exit_SUCCESS;
+	} while (ok && ++xoptind < argc);
+	return ok ? Exit_SUCCESS : Exit_FAILURE;
     }
 
     bool err = false;
@@ -1077,10 +1083,15 @@ set_traps:
     return err ? Exit_FAILURE : Exit_SUCCESS;
 
 print_usage:
-    fprintf(stderr,
-	    Ngt("Usage:  trap [action signal...]\n"
-		"        trap signum [signal...]\n"
-		"        trap -p [signal...]\n"));
+    if (posixly_correct)
+	fprintf(stderr,
+		Ngt("Usage:  trap [action signal...]\n"
+		    "        trap signum [signal...]\n"));
+    else
+	fprintf(stderr,
+		Ngt("Usage:  trap [action signal...]\n"
+		    "        trap signum [signal...]\n"
+		    "        trap -p [signal...]\n"));
     SPECIAL_BI_ERROR;
     return Exit_ERROR;
 }
@@ -1088,14 +1099,20 @@ print_usage:
 /* Prints a trap command to stdout that can be used to restore the current
  * signal handler for the specified signal.
  * If the `command' is NULL, this function does nothing.
- * Otherwise, the `command' is properly single-quoted and printed. */
-void print_trap(const char *signame, const wchar_t *command)
+ * Otherwise, the `command' is properly single-quoted and printed.
+ * Returns true iff successful (no error). */
+bool print_trap(const char *signame, const wchar_t *command)
 {
     if (command) {
 	wchar_t *q = quote_sq(command);
-	printf("trap -- %ls %s\n", q, signame);
+	int r = printf("trap -- %ls %s\n", q, signame);
+	if (r < 0)
+	    xerror(errno, Ngt("cannot print to standard output"));
 	free(q);
+	if (r < 0)
+	    return false;
     }
+    return true;
 }
 
 const char trap_help[] = Ngt(
@@ -1112,7 +1129,7 @@ const char trap_help[] = Ngt(
 "If the first operand is a non-negative integer, the operand is considered as\n"
 "a signal specification and <action> is assumed to be \"-\".\n"
 "If the -p (--print) option is specified, the actions for the specified\n"
-"<signal>s are printed.\n"
+"<signal>s are printed. This option is not available in POSIXly correct mode.\n"
 "Without any operands, all signal handlers currently set are printed.\n"
 );
 
@@ -1123,10 +1140,8 @@ const char trap_help[] = Ngt(
  * -v: prints signal info verbosely */
 int kill_builtin(int argc, void **argv)
 {
-    if (!posixly_correct && argc == 2 && wcscmp(ARGV(1), L"--help") == 0) {
-	print_builtin_help(ARGV(0));
-	return Exit_SUCCESS;
-    }
+    if (!posixly_correct && argc == 2 && wcscmp(ARGV(1), L"--help") == 0)
+	return print_builtin_help(ARGV(0));
 
     wchar_t opt;
     int signum = SIGTERM;
