@@ -45,7 +45,7 @@ extern void ignore_sigtstp(void);
 extern _Bool send_sigstop_to_myself(void);
 
 extern int wait_for_sigchld(_Bool interruptible, _Bool return_on_trap);
-extern void wait_for_input(int fd, _Bool trap);
+extern _Bool wait_for_input(int fd, _Bool trap, int timeout);
 
 extern void handle_sigchld(void);
 extern _Bool have_unhandled_traps(void);
@@ -60,7 +60,8 @@ extern int kill_builtin(int argc, void **argv)
     __attribute__((nonnull));
 extern const char trap_help[], kill_help[];
 
-#if HAVE_STRSIGNAL
+#if HAVE_STRSIGNAL && _POSIX_C_SOURCE < 200809L
+# undef strsignal
 extern char *strsignal(int signum);
 #endif
 
@@ -68,4 +69,4 @@ extern char *strsignal(int signum);
 #endif /* YASH_SIG_H */
 
 
-/* vim: set ts=8 sts=4 sw=4 noet: */
+/* vim: set ts=8 sts=4 sw=4 noet tw=80: */
