@@ -25,46 +25,50 @@
 extern char **environ;
 
 /* variable names */
-#define VAR_CDPATH	"CDPATH"
-#define VAR_COLUMNS	"COLUMNS"
-#define VAR_DIRSTACK	"DIRSTACK"
-#define VAR_ENV		"ENV"
-#define VAR_FCEDIT	"FCEDIT"
-#define VAR_HISTFILE	"HISTFILE"
-#define VAR_HISTRMDUP   "HISTRMDUP"
-#define VAR_HISTSIZE	"HISTSIZE"
-#define VAR_HOME	"HOME"
-#define VAR_IFS		"IFS"
-#define VAR_LANG	"LANG"
-#define VAR_LC_ALL	"LC_ALL"
-#define VAR_LC_COLLATE	"LC_COLLATE"
-#define VAR_LC_CTYPE	"LC_CTYPE"
-#define VAR_LC_MESSAGES	"LC_MESSAGES"
-#define VAR_LC_MONETARY	"LC_MONETARY"
-#define VAR_LC_NUMERIC	"LC_NUMERIC"
-#define VAR_LC_TIME	"LC_TIME"
-#define VAR_LINENO	"LINENO"
-#define VAR_LINES	"LINES"
-#define VAR_MAIL	"MAIL"
-#define VAR_MAILCHECK	"MAILCHECK"
-#define VAR_MAILPATH	"MAILPATH"
-#define VAR_NLSPATH	"NLSPATH"
-#define VAR_OLDPWD	"OLDPWD"
-#define VAR_OPTARG	"OPTARG"
-#define VAR_OPTIND	"OPTIND"
-#define VAR_PATH	"PATH"
-#define VAR_PPID	"PPID"
-#define VAR_PROMPT_COMMAND "PROMPT_COMMAND"
-#define VAR_PS1		"PS1"
-#define VAR_PS2		"PS2"
-#define VAR_PS3		"PS3"
-#define VAR_PS4		"PS4"
-#define VAR_PWD		"PWD"
-#define VAR_RANDOM	"RANDOM"
-#define VAR_TERM	"TERM"
-#define VAR_YASH_LE_TIMEOUT "YASH_LE_TIMEOUT"
-#define VAR_YASH_VERSION "YASH_VERSION"
-#define L               L""
+#define VAR_CDPATH                    "CDPATH"
+#define VAR_COLUMNS                   "COLUMNS"
+#define VAR_COMMAND_NOT_FOUND_HANDLER "COMMAND_NOT_FOUND_HANDLER"
+#define VAR_DIRSTACK                  "DIRSTACK"
+#define VAR_ECHO_STYLE                "ECHO_STYLE"
+#define VAR_ENV                       "ENV"
+#define VAR_FCEDIT                    "FCEDIT"
+#define VAR_HANDLED                   "HANDLED"
+#define VAR_HISTFILE                  "HISTFILE"
+#define VAR_HISTRMDUP                 "HISTRMDUP"
+#define VAR_HISTSIZE                  "HISTSIZE"
+#define VAR_HOME                      "HOME"
+#define VAR_IFS                       "IFS"
+#define VAR_LANG                      "LANG"
+#define VAR_LC_ALL                    "LC_ALL"
+#define VAR_LC_COLLATE                "LC_COLLATE"
+#define VAR_LC_CTYPE                  "LC_CTYPE"
+#define VAR_LC_MESSAGES               "LC_MESSAGES"
+#define VAR_LC_MONETARY               "LC_MONETARY"
+#define VAR_LC_NUMERIC                "LC_NUMERIC"
+#define VAR_LC_TIME                   "LC_TIME"
+#define VAR_LINENO                    "LINENO"
+#define VAR_LINES                     "LINES"
+#define VAR_MAIL                      "MAIL"
+#define VAR_MAILCHECK                 "MAILCHECK"
+#define VAR_MAILPATH                  "MAILPATH"
+#define VAR_NLSPATH                   "NLSPATH"
+#define VAR_OLDPWD                    "OLDPWD"
+#define VAR_OPTARG                    "OPTARG"
+#define VAR_OPTIND                    "OPTIND"
+#define VAR_PATH                      "PATH"
+#define VAR_PPID                      "PPID"
+#define VAR_PROMPT_COMMAND            "PROMPT_COMMAND"
+#define VAR_PS1                       "PS1"
+#define VAR_PS2                       "PS2"
+#define VAR_PS3                       "PS3"
+#define VAR_PS4                       "PS4"
+#define VAR_PWD                       "PWD"
+#define VAR_RANDOM                    "RANDOM"
+#define VAR_TERM                      "TERM"
+#define VAR_YASH_AFTER_CD             "YASH_AFTER_CD"
+#define VAR_YASH_LE_TIMEOUT           "YASH_LE_TIMEOUT"
+#define VAR_YASH_VERSION              "YASH_VERSION"
+#define L                             L""
 
 typedef enum path_T {
     PA_PATH, PA_CDPATH, PA_MAILPATH,
