@@ -261,7 +261,7 @@ int main(int argc, char **argv)
 
 #if YASH_ENABLE_LINEEDIT
     /* enable lineedit if interactive and connected to a terminal */
-    if (is_interactive)
+    if (is_interactive && shopt_lineedit == shopt_nolineedit)
 	if (!is_interactive_set
 		|| (isatty(STDIN_FILENO) && isatty(STDERR_FILENO)))
 	    shopt_lineedit = shopt_vi;
@@ -579,14 +579,18 @@ void parse_and_exec(parseinfo_T *pinfo, bool finally_exit)
     for (;;) {
 	and_or_T *commands;
 
-	forceexit = nextforceexit;
-	nextforceexit = false;
+	if (pinfo->intrinput) {
+	    forceexit = nextforceexit;
+	    nextforceexit = false;
+	}
 	if (return_pending()) {
-	    reset_execinfo();
 	    if (pinfo->intrinput) {
+		reset_execinfo();
 		xerror(0, Ngt("return: not in function or sourced file"));
+		laststatus = Exit_FAILURE;
 	    } else if (finally_exit) {
 		wchar_t argv0[] = L"return";
+		reset_execinfo();
 		exit_builtin(1, (void *[]) { argv0, NULL });
 		continue;
 	    } else
@@ -669,8 +673,7 @@ int exit_builtin(int argc, void **argv)
 		forceexit = true;
 		break;
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:  print_usage:
 		fprintf(stderr, gt("Usage:  exit [-f] [n]\n"));
 		SPECIAL_BI_ERROR;
@@ -689,18 +692,21 @@ int exit_builtin(int argc, void **argv)
 		    sjc),
 		sjc);
 	fprintf(stderr, gt("  Use `exit' again to exit anyway.\n"));
-	nextforceexit = true;
+	forceexit = nextforceexit = true;
 	return Exit_FAILURE;
     }
 
     int status;
     const wchar_t *statusstr = ARGV(xoptind);
-    if (statusstr != NULL
-	    && xwcstoi(statusstr, 10, &status)
-	    && status >= 0) {
-	status &= 0xFF;
+    if (statusstr != NULL) {
+	if (xwcstoi(statusstr, 10, &status) && status >= 0) {
+	    status &= 0xFF;
+	} else {
+	    xerror(0, Ngt("`%ls' is not a valid integer"), statusstr);
+	    status = Exit_ERROR;
+	    SPECIAL_BI_ERROR;
+	}
     } else {
-	/* ignore `statusstr' if not a valid non-negative integer */
 	status = -1;
     }
     exit_shell_with_status(status);
@@ -732,8 +738,7 @@ int suspend_builtin(int argc, void **argv)
 	    case L'f':
 		break;
 	    case L'-':
-		print_builtin_help(ARGV(0));
-		return Exit_SUCCESS;
+		return print_builtin_help(ARGV(0));
 	    default:  print_usage:
 		fprintf(stderr, gt("Usage:  suspend\n"));
 		return Exit_ERROR;
